package fileops;

use constant OS=>$^O!~/MS|Win|Windows|Win32|Microsoft/o;

sub osopen  {
	my($file, $mode, $fh)=(shift, shift || 4);
	open $fh, $file and binmode $fh or die $file, "\n", $!;
	+$fh
}

sub osclose {
	my($fh, $mode)=(shift, shift || 8);
	+ close $fh or die $!
}

sub osreadfile {
	my($file, $fh)=shift;
	$fh=&osopen($file);
		my$old=$/;$/=undef;
		my$str=<$fh>;$/=$old;
	&osclose($fh);
	+$str
}

sub oswritefile {
	my($file, $fh)=shift;
	+ $fh=&osopen('>'.$file, 2) and print $fh @_ and &osclose($fh)
}

package RXParser;

sub new {
	my($class, $self)=(shift, {@_});
	+bless $self, $class
}

sub save {
	my($self, $file)=(shift, shift);
	use Data::Dumper;($Data::Dumper::Terse, $Data::Dumper::Indent)=1;
	&fileops::oswritefile($file, '+ ', Dumper $self);
	+$self
}

sub load {+do $_[1]}

sub error {
	my$self=shift;
	print STDERR "\n", (caller)[0], ': ', @_, ' at ',
				(caller)[1], ' line ', (caller)[1, 2], "\n" unless $self->{silent};
	exit;
}

package RXParser::Filter;
use Entities qw'decode_entities';

BEGIN {
	%MODE=(
		'HTML'=>sub{+&decode_entities(shift)},
		'TEXT'=>sub{+&decode_entities($MODE{'TEXT/ENTITY'}(shift))},
		'TEXT/ENTITY'=>sub{
			my$str=shift;
			$str=~s/<[^>]+>//gos;
			+$str
		},
		'DATE'=>sub{
			my$date=shift;
			return $date unless $date=~/(\d{1,2})\D+(\d{1,2})\D+(\d{4}|\d{2})/o;
			+sprintf '%02d.%02d.%04d', $1, $2, $3<40?$3+2000:$3
		},
		'LINK'=>sub{+$MODE{'TEXT/ENTITY'}(shift)}
	);
}

sub prepare {
	my($self, $str, $mode)=(shift, shift, uc shift);
	+ ref $MODE{$mode} eq 'CODE' ? $MODE{$mode}($str) : $str
}

sub trim {+$_[0]=~/^\s*(.*?)\s*$/os? $1 : $_[0]}

package RXParser::Mapping;
use base 'RXParser';

sub prepare {
	my$self=shift;
	my($vi, $vp)=(1, 'v%d');
	my$filename=shift || $self->{file} || $self->error('Input rules filename not found');
	my$fh=&fileops::osopen($filename);
	$self->{result}={};

		while(<$fh>) {
			next unless index $_, '#';
			if(/^\s*(['"]?)(.+?)\1\s*(?:\:\s*(.+?))?\s*\((.+?)\)\s*(.)[\r\n]*$/os) {
				my($title, $name, $args, $dot, $i, %args)=($2, $3, $4, qr(^\Q$5\E\s*$)o, 0);
				$name||=sprintf $vp, $vi;
				while($args=~/\s*(["']?)([^\1,\:]+)\1\s*(?:\:\s*(["']?)([^,]+)\3)?\s*,?/goc) {
					my$key=&RXParser::Filter::trim($2);
					$args{$key}={
						order=>++$i,
						type=>$4
					} unless $key eq undef
				}
				$self->{result}{$name}={
					order=>$vi++,
					title=>$title,
					args=>{%args}
				};
				while(<$fh>) {
					last if /$dot/o;
					s/^\s*(.*?)\s*$/\\s\*$1\\s\*/o;
					$self->{result}{$name}{rx}.=$_;
				}
				$self->{result}{$name}{rx}=~s/(?:\\s\*){2,}/\\s*/go;
				$self->{result}{$name}{rx}=~s/^(?:\\s\*)(.*?)(?:\\s\*)$/$1/go;
				$self->{result}{$name}{rx}=qr($self->{result}{$name}{rx})si;
			}
		}
	&fileops::osclose($fh);
	+$self
}

+1;

package RXParser::Parser;
use base 'RXParser';

sub prepare {
	my $self=shift;
	$self->{mapping}=shift  || $self->{mapping}
					or $self->error('Input mapping not found');
	$self->{file}=shift || $self->{file}
					or $self->error('Input filename not found');
	my $file=&fileops::osreadfile($self->{file});study $file;
	my($q, $mr, $i, $key, $res, @keys)=(0, $self->{mapping}{result});
	$self->{result}={};

	foreach $key (sort{$mr->{$a}{order}<=>$mr->{$b}{order}}keys %$mr) {
		$self->{result}{$key}={
			found=>[],
			info=>$mr->{$key}
		};
		$q=0;
		@keys=sort{$mr->{$key}{args}{$a}{order}<=>$mr->{$key}{args}{$b}{order}}
								keys%{$mr->{$key}{args}};
		pos($file)=0;
		while($file=~/$mr->{$key}{rx}/gc) {
			$res={
				'#'=>+$q++,
				'!'=>+pos $file
			};
			for($i=0; $i<@keys; $i++) {
				$res->{$keys[$i]}=RXParser::Filter->prepare(${$i+1},
						$mr->{$key}{args}{$keys[$i]}{type})
			}
			push @{$self->{result}{$key}{found}}, $res
		}
	}
	+$self
}

+1;

package RXParser::RXML;
use base 'RXParser';

BEGIN {
	$PREFIX='rxml';
	%PREFIXES=(
		'iter'=>sub{
				my($self, $tree, $parent, $index, $parser, $i)=(
						shift, shift, shift, shift, shift, 0);
				my$cur=$parser->{$tree->{node}}{found};
				my$len=scalar @$cur;

				if(ref $parent eq 'HASH' && $parent->{prefix} eq 'iter') {
					my$par=$parser->{$parent->{node}}{found};

					if($index==$#$par) {
						for(; $i<$len; $i++) {
							$self->buildout($tree->{content}, $tree, $i)
									if $par->[$index]{'!'} < $cur->[$i]{'!'}
						}
					} else {
						for(; $i<$len; $i++) {
							$self->buildout($tree->{content}, $tree, $i)
									if $par->[$index]{'!'} < $cur->[$i]{'!'} &&
										$par->[$index+1]{'!'} > $cur->[$i]{'!'}
						}
					}
				} else {
					for(; $i<$len; $i++) {
						$self->buildout($tree->{content}, $tree, $i)
					}
				}

				+1
		}
	);
}

sub explode {
	my$self=shift;
	+[$self->buildtree(&fileops::osreadfile(shift))]
}

sub buildnode {
	my$self=shift;
	my($str, $pos, @result)=shift;
	if($str=~s/<$PREFIX\:(["']?)([^>]+?)\1\s+([^>]+?)\s*(?:\/>|>\s*(.*?)\s*<\/$PREFIX\:(["']?)\2\5>)/
		push @result, substr($`, $pos, pos($str)-$pos), {
			content=>$4,
			node=>$2,
			params=>$self->getparams($3)
		};
		$pos=pos($str)+length($&);
		+undef
	/geios) {
		+ @result, $'
	} else {
		+ $str
	}
}

sub buildtree {
	my$self=shift;
	my($str, $pos, @result, $prefix, $content, $node, $params)=shift;
	my$len=length $str;
	if($str=~s/<$PREFIX(\-\w+)\:(["']?)([^>]+?)\2(?:([^>]+?))?>\s*(.*?)\s*<\/$PREFIX\1\:(["']?)\3\6>/
		if(exists $PREFIXES{+$prefix=lc substr $1, 1}) {
			($content, $node, $params)=($5, $3, $4);
			push @result, $self->buildnode(substr $`, $pos, pos($str)-$pos),
			{
				prefix=>$prefix,
				content=>$content,
				node=>$node,
				params=>$self->getparams($params)
			};
		}
		$pos=pos($str)+length($&);
		+undef
	/geios) {
		push @result, $';
		foreach(@result) {
			if(ref eq 'HASH') {
				if(defined $_->{content}) {
					$_->{content}=[$self->buildtree($_->{content})]
				} else {
					delete $_->{content}
				}
			} else {
				$_=[$self->buildnode($_)]
			}
		}
		+ @result
	} else {
		+ $self->buildnode($str)
	}
}

sub buildout {
	my($self, $tree, $parent, $index)=(shift, shift, shift, shift);

	if(ref $tree eq 'ARRAY') {
		$self->buildout($_, $parent, $index) foreach @$tree;
	} elsif(ref $tree eq 'HASH') {

		my$parser=$self->{parser}{result};
		if(exists $parser->{$tree->{node}}) {
			if(ref $PREFIXES{$tree->{prefix}} eq 'CODE') {
				$PREFIXES{$tree->{prefix}}($self, $tree, $parent, $index, $parser);
			} else {
				$index=$tree->{params}{index} if length $tree->{params}{index};
				print RXParser::Filter->prepare(
					$parser->{$tree->{node}}{found}[$index]{$tree->{params}{value}},
					$tree->{params}{type}
				);
			}
		} else {
			print "\n", '<!--', $tree->{node}, '-->', "\n"
		}

	} else {
		print $tree
	}
}

sub prepare {
	my$self=shift;
	my$file=$self->{file} or $self->error('Input template filename not found');
	$self->{tree}=$self->explode($file);
	+$self
}

sub saveAs {

	#     

	my$self=shift;
	return undef unless $self->{tree};
	my$parser=$self->{parser} or $self->error('RXParser::Parser object is not defined');
	my$filename=shift || $self->{output} || $self->error('Output filename not found');
	my$fh=&fileops::osopen('>'.$filename, 2);
	my$old=select $fh;$self->buildout($self->{tree});select $old;
	+&fileops::osclose($fh)
}

sub getparams {
	#     :
	# name1="value1" name2="value2" name3="value3"...

	my($self, $str, %result)=(shift, shift);
	$str=~s/(\w+)\s*=\s*(["']?)(.*?)\2/$result{+lc $1}=$3; +undef/geos;
	+\%result
}

+1;

__END__
=head1 NAME

 RXParser 1.7125 - Universal document translator

=head1 AUTHOR

 Shatrow Alexey <cdch@yandex.ru>,
 Krasnodar, March 13, 2005 (c)
=cut
